//
// © 2005-2008 Marko Lindqvist
//

#include "sudoku.h"

#include <assert.h>
#include <iostream>
#include <sys/time.h>
#include <time.h>

#include <sstream>

#include "cell.h"
#include "sdksoption.h"
#include "sdksui.h"
#include "sdksinput.h"


sudoku::sudoku(sdks_input *input_in)
{
  sudoku(input_in, 0);
}

sudoku::sudoku(sdks_input *input_in, sdks_ui *ui_in)
{
  int i, j;

  input = input_in;

  if (!ui_in) {
    ui = new sdks_ui;
    ui_alloc = true;
  } else {
    ui = ui_in;
    ui_alloc = false;
  }

  base_size = input->get_size();
  dimension = base_size*base_size;
  unsolved = dimension*dimension;
  cells = new cell**[dimension];

  for (i = 0; i < dimension; i++) {
    cells[i] = new cell*[dimension];
    for (j = 0; j < dimension; j++) {
      cells[i][j] = new cell(this);
    }
  }
}


sudoku::~sudoku()
{
  int i, j;

  for (i = 0; i < dimension ; i++) {
    for (j = 0; j < dimension; j++) {
      delete cells[i][j];
    }
    delete cells[i];
  }

  delete cells;

  if (ui_alloc) {
    // sudoku allocated its own printer
    delete ui;
  }
}


int sudoku::get_size()
{
  return base_size;
}


sdks_ui *sudoku::get_ui()
{
  return ui;
}


bool sudoku::check_legality()
{
  int row, col, reg, linear, i;
  bool legal = true;
  int result;
  char logmsg[50];
  cell *pcell, *ccell;

  // Rows
  for (row = 0; row < dimension; row++) {
    for (col = 0; col < dimension; col++) {
      pcell = get_cell(CT_ROW, row, col);
      result = pcell->solved();
      if (result) {
        for (i = col + 1; i < dimension; i++) {
          ccell = get_cell(CT_ROW, row, i);
          if (ccell->solved() == result) {
            // Two cells in same row has same result
            sprintf(logmsg, "Row has %d also in (%d,%d)", result, i, row);
            ui->solve_log(pcell, logmsg);

            legal = false;
          }
        }
      }
    }
  }

  // Columns
  for (col = 0; col < dimension; col++) {
    for (row = 0; row < dimension; row++) {
      pcell = get_cell(CT_COL, col, row);
      result = pcell->solved();
      if (result) {
        for (i = row + 1; i < dimension; i++) {
          ccell = get_cell(CT_COL, col, i);
          if (ccell->solved() == result) {
            // Two cells in same column has same result
            sprintf(logmsg, "Column has %d also in (%d,%d)", result, col, i);
            ui->solve_log(pcell, logmsg);

            legal = false;
          }
        }
      }
    }
  }

  // Regions
  for (reg = 0; reg < dimension; reg++) {
    for (linear = 0; linear < dimension; linear++) {
      pcell = get_cell(CT_REG, reg, linear);
      result = pcell->solved();
      if (result) {
        for (i = linear + 1; i < dimension; i++) {
          ccell = get_cell(CT_REG, reg, i);
          if (ccell->solved() == result) {
            // Two cells in same region has same result
            sprintf(logmsg, "Region has %d also in (%d,%d)", result,
                    linear_to_x(reg, i), linear_to_y(reg, i));
            ui->solve_log(pcell, logmsg);

            legal = false;
          }
        }
      }
    }
  }

  return legal;
}


bool sudoku::read()
{
  int i, j;
  bool ok;

  ok = input->is_ok();

  for(i = 0; i < dimension; i++) {
    for(j = 0; j < dimension; j++) {
      int value;

      get_row_cell(i, j)->setxy(j, i);
      value = input->get_value(i, j);

      if (value != -1) {
        get_row_cell(i, j)->set_result(value, "Preset value");
        unsolved--;
      }
    }
  }

  return ok;
}


bool sudoku::solve()
{
  int method;
  bool advanced = true;
  bool first_round = true;
  struct timeval start;
  struct timeval end;
  bool solvetime = get_option(OPTION_SOLVETIME);

  if (solvetime) {
    gettimeofday(&start, NULL);
  }

  while (advanced && unsolved) {
    advanced = false;

    if (!first_round && get_option(OPTION_INTERMED)) {
      ui->print_sudoku_possibilities(this);
    }
    first_round = false;

    for (method = METHOD_BASE ; !advanced && method < METHOD_LAST; method++) {
      advanced = apply_solving_method((solving_method)method);
    }
  }


  if (solvetime) {
    gettimeofday(&end, NULL);

    int secdiff = end.tv_sec - start.tv_sec;
    int usecdiff = end.tv_usec - start.tv_usec;
    int milldiff;

    if (usecdiff < 0) {
      secdiff++;
      usecdiff -= 1000000;
    }

    milldiff = usecdiff / 1000;
    usecdiff -= milldiff * 1000;

    std::cout << "Completed in " << secdiff << " seconds and " << milldiff << "." << usecdiff << " milliseconds." << std::endl;
  }

  if(unsolved) {
    std::string numstr;
    std::stringstream numstream(numstr, std::ios_base::out);
    std::string unsmsg;

    numstream << unsolved;
    unsmsg = "Unsolvable (" + numstream.str() + " left)";

    ui->print_msg(unsmsg);
    if (get_option(OPTION_INTERMED)) {
      ui->print_sudoku_possibilities(this);
    }
    return false;
  }
  return true;
}


bool sudoku::apply_solving_method(solving_method method)
{
  bool advanced = false;
  int i;

  switch(method) {
   case METHOD_BASE:
     if (get_option(OPTION_INTERMED)) {
       ui->print_method_name("Base");
     }
     for(i = 0; i < dimension; i++) {
       advanced |= solve_base(CT_ROW, i);
     }
     for(i = 0; i < dimension; i++) {
       advanced |= solve_base(CT_COL, i);
     }
     for(i = 0; i < dimension; i++) {
       advanced |= solve_base(CT_REG, i);
     }
     break;
   case METHOD_ROWCOL_FOR_REGION:
     if (get_option(OPTION_INTERMED)) {
       ui->print_method_name("Rowcol for region");
     }
     for(i = 0; i < dimension; i++) {
       advanced |= solve_region_slow(i);
     }
     break;
   case METHOD_REGION_FOR_ROWCOL:
     if (get_option(OPTION_INTERMED)) {
       ui->print_method_name("Region for rowcol");
     }
     for(i = 0; i < dimension; i++) {
       advanced |= solve_slow_rowcol(CT_ROW, i);
     }
     for(i = 0; i < dimension; i++) {
       advanced |= solve_slow_rowcol(CT_COL, i);
     }
     break;
   case METHOD_PAIRING:
     if (get_option(OPTION_INTERMED)) {
       ui->print_method_name("Pairing");
     }
     for (i = 0; i < dimension; i++) {
       advanced |= solve_pairing(CT_ROW, i);
     }
     for (i = 0; i < dimension; i++) {
       advanced |= solve_pairing(CT_COL, i);
     }
     for (i = 0; i < dimension; i++) {
       advanced |= solve_pairing(CT_REG, i);
     }
     break;
   case METHOD_REVEAL_PAIRING:
#ifdef METHOD_REVEAL_PAIRING
     if (get_option(OPTION_INTERMED)) {
       ui->print_method_name("Reveal pairing");
     }
     for (i = 0; i < dimension; i++) {
       advanced |= solve_reveal_pairing(CT_ROW, i);
     }
     for (i = 0; i < dimension; i++) {
       advanced |= solve_reveal_pairing(CT_COL, i);
     }
     for (i = 0; i < dimension; i++) {
       advanced |= solve_reveal_pairing(CT_REG, i);
     }
#endif // METHOD_REVEAL_PAIRING
     break;
   case METHOD_LAST:
     assert(false);
     break;
  }

  return advanced;
}


bool sudoku::solve_base(construct_type construct, int cnum)
{
  int i, j;
  bool advanced = false;
  char msg_already[50];
  char msg_only[50];

  sprintf(msg_already, "Value already in %s", construct_name(construct));
  sprintf(msg_only, "Only place for value in %s", construct_name(construct));

  for(i = 0; i < dimension; i++) {
    int cleanup = get_cell(construct, cnum, i)->solved();
    if(cleanup) {
      for(j = 0; j < dimension; j++) {
        if(i != j) {
          remove_result rr =
            get_cell(construct, cnum, j)->remove_possibility(cleanup,
                                                             msg_already);
          if(rr != RR_KNOWN) {
            advanced = true;
            if(rr == RR_SOLVED) {
              unsolved--;
            }
          }
        }
      }
    }
  }
  for(i = 1; i <= dimension; i++) {
    int place = -1;
    for(j = 0; j < dimension; j++) {
      if(get_cell(construct, cnum, j)->is_possible(i)) {
        if(place == -1) {
          place = j;
        } else {
          place = -2;
        }
      }
    }
    if(place >= 0) {
      if(get_cell(construct, cnum, place)->set_result(i, msg_only)) {
        advanced = true;
        unsolved--;
      }
    }
  }
  return advanced;
}


bool sudoku::solve_slow_rowcol(construct_type construct, int cnum)
{
  int i, j;
  int reg;
  bool advanced = false;

  assert(construct != CT_REG);

  for(i = 1; i <= dimension; i++) { // Each possible cell value
    int only_region = -1;

    // Whether value for this row/col can be in region
    for (reg = 0; reg < base_size; reg++) {
      bool result;
      int abs_reg = get_reg(construct, cnum, reg);

      for (j = 0; j < base_size; j++) {
        if (construct == CT_ROW) {
          result = get_row_cell(cnum, region_origo_x(abs_reg) + j); 
        } else {
          assert(construct == CT_COL);
          result = get_col_cell(cnum, region_origo_y(abs_reg) + j);
        }
        if (result) {
          if (only_region != -1 && only_region != abs_reg) {
            // Already had region, so there is several of them
            only_region = -2;
            break;
          } else {
            only_region = abs_reg;
          }
        }
      }
      if (only_region == -2) {
        break;
      }
    }

    if (only_region >= 0) {
      // Value for this row/col must be in region only_region
      char logmsg[50];

      sprintf(logmsg, "At %s %d, value can be only at this region (%d)",
              construct_name(construct), cnum, only_region);

      for (j = 0; j < dimension; j++) {
        if ((construct == CT_ROW && cnum != linear_to_y(only_region, j)) ||
            (construct == CT_COL && cnum != linear_to_x(only_region, j))) {
          // Linear coordinate j is not in this row/col
          remove_result rr = get_reg_cell(only_region, j)->remove_possibility(i, logmsg);
          if(rr != RR_KNOWN) {
            advanced = true;
            if(rr == RR_SOLVED) {
              unsolved--;
            }
          }
        }
      }
    }

  }
  return advanced;
}


bool sudoku::solve_region_slow(int region)
{
  int i, j;
  int row, col; // Relative to region
  bool advanced = false;

  for(i = 1; i <= dimension; i++) { // Each possible cell value
    int only_row = -1;
    int only_col = -1;

    // Whether value for this region can be in row/column
    for (row = 0; row < base_size; row++) {
      for (col = 0; col < base_size; col++) {
        bool result;
        result = get_regxy_cell(region, col, row)->is_possible(i);
        if (result) {
          if (only_row != -1 && only_row != row) {
            // Already had row, so there is several of them.
            only_row = -2;
          } else {
            only_row = row;
          }
          if (only_col != -1 && only_col != col) {
            // Already had col, so there is several of them.
            only_col = -2;
          } else {
            only_col = col;
          }
        }
      }
    }

    if (only_row >= 0) {
      // Value for this region must be in row only_row

      char rowmsg[50];
      sprintf(rowmsg, "At region %d, value can be only at this row (%d)",
              region, only_row);

      for (j = 0; j < dimension; j++) {
        if (!region_x(region, j)) {
          // X-coordinate j is in another region
          remove_result rr = get_row_cell(region_origo_y(region) + only_row, j)->
            remove_possibility(i, rowmsg);
          if(rr != RR_KNOWN) {
            advanced = true;
            if(rr == RR_SOLVED) {
              unsolved--;
            }
          }
        }
      }
    }

    if (only_col >= 0) {
      // Value for this region must be in col only_col

      char colmsg[50];
      sprintf(colmsg, "At region %d, value can be only at this column (%d)",
              region, only_col);

      for (j = 0; j < dimension; j++) {
        if (!region_y(region, j)) {
          // Y-coordinate j is in another region
          remove_result rr = get_col_cell(region_origo_x(region) + only_col, j)->
            remove_possibility(i, colmsg);
          if(rr != RR_KNOWN) {
            advanced = true;
            if(rr == RR_SOLVED) {
              unsolved--;
            }
          }
        }
      }
    }

  }
  return advanced;
}


bool sudoku::solve_pairing(construct_type construct, int cnum)
{
  bool advanced = false;
  int start;                         // Cell we search pairings for
  int i, j, pos, num, positions;
  bool possibilities[dimension + 1]; // Values in perfect pairing
  bool pairings[dimension];          // Whether cell in given index is part of pairing
  cell *pcell;
  char logmsg[50];

  sprintf(logmsg, "Pairing in %s", construct_name(construct));

  for (start = 0; start < dimension; start++) {
    pcell = get_cell(construct, cnum, start);
    if (!pcell->solved()) {
      // There can be pairing only with unsolved cells

      // Empty pairings array
      for (i = 0; i < dimension; i++) {
        pairings[i] = false;
      }
      pairings[start] = true;

      num = 0;
      positions = 1;

      // Fill possibilities array with values from start cell
      for (pos = 1; pos <= dimension; pos++ ) {
        if (pcell->is_possible(pos)) {
          possibilities[pos] = true;
          num++;
        } else {
          possibilities[pos] = false;
        }
      }

      // Iterate through all the other cells in construct.
      // We cannot just start 'i = start + 1', since checking pairing
      // between cells A->B may differ from check B->A if cells have
      // different number of possibilities left
      for (i = 0; i < dimension; i++) {
        if (start != i) {
          cell *ccell = get_cell(construct, cnum, i); // Start cell is compared to this
          bool same_pos = true;

          for (pos = 1; pos <= dimension; pos++) {
            if (!possibilities[pos] && ccell->is_possible(pos)) {
              // It's ok that compared cell does not have some possibility
              // that start cell has.
              same_pos = false;                       // Possibilites differ
              break;
            }
          }
          if (same_pos) {
            // This cell pairs with start cell
            positions++;
            pairings[i] = true;
            if (num == positions) {
              // There is strict pairing
              // These possibilities are required to fill pairings cells, remove them
              // from other cells.
              for (j = 0; j < dimension; j++) {
                ccell = get_cell(construct, cnum, j);
                if (!pairings[j] && !ccell->solved()) {
                  // Cell is not part of pairing and not solved
                  for (pos = 1; pos <= dimension; pos++) {
                    if (possibilities[pos]) {
                      // This possibility is part of pairing, remove from this cell
                      remove_result rr;
                      
                      rr = ccell->remove_possibility(pos, logmsg);
                      if (rr != RR_KNOWN) {
                        advanced = true;
                        if(rr == RR_SOLVED) {
                          unsolved--;
                        }
                      }
                    }
                  }
                }
              }
            }
          } // same_pos
        } // start != i
      }
    }
  }

  return advanced;
}


bool sudoku::solve_reveal_pairing(construct_type construct, int cnum)
{
  return false;

  bool advanced = false;
  int start;                         // Cell we search pairings for
  int i, j, pos, num, positions;
  bool possibilities[dimension + 1]; // Values in perfect pairing
  bool pairings[dimension];          // Whether cell in given index is part of pairing
  cell *pcell;
  char logmsg[50];

  sprintf(logmsg, "Revealed pairing in %s", construct_name(construct));

  for (start = 0; start < dimension; start++) {
    pcell = get_cell(construct, cnum, start);
    if (!pcell->solved()) {
      // There can be pairing only with unsolved cells

      int posarray[dimension + 1];
      bool posarrayinuse[dimension + 1];
      int posarraysize = 0;

      // Fill posarray from start cell
      for (pos = 1; pos <= dimension; pos++ ) {
        if (pcell->is_possible(pos)) {
          posarray[posarraysize] = pos;
          posarrayinuse[posarraysize++] = false;
        }
      }

      if (posarraysize > 2) {
        // Cell with less than 3 possibilites cannot contain possibility pair *and*
        // extra possibilities. We are in the process of removing extra possibilities

        bool new_set_found = true;

        // Empty possibilities array
        for (pos = 1; pos <= dimension; pos++) {
          possibilities[pos] = false;
        }

        // Initially two first possibilities set
        possibilities[posarray[0]] = true;
        posarrayinuse[0] = true;
        possibilities[posarray[1]] = true;
        posarrayinuse[1] = true;
        num = 2;

        while(new_set_found) {
          cell *ccell;

          // Empty pairings array
          for (i = 0; i < dimension; i++) {
            pairings[i] = false;
          }
          pairings[start] = true;
          positions = 1;

          // Iterate through all the other cells in construct.
          // We cannot just start 'i = start + 1', since checking pairing
          // between cells A->B may differ from check B->A if cells have
          // different number of possibilities left
          for (i = 0; i < dimension; i++) {
            if (start != i) {
              ccell = get_cell(construct, cnum, i); // Start cell is compared to this
              bool same_pos = false;

              for (pos = 1; pos <= dimension; pos++) {
                if (possibilities[pos] && ccell->is_possible(pos)) {
                  // Compared cell has some same possibilities than start cell
                  same_pos = false;                       // Possibilites differ
                  break;
                }
              }
              if (same_pos) {
                // This cell pairs with start cell
                positions++;
                pairings[i] = true;
              }
            }
          }

          if (num == positions) {
            // There is strict pairing
            // Pairing cells are required for pairing values
            for (j = 0; j < dimension; j++) {
              ccell = get_cell(construct, cnum, j);
              if (pairings[j]) {
                // Cell is part of pairing
                for (pos = 1; pos <= dimension; pos++) {
                  if (!possibilities[pos]) {
                    // This possibility is not part of pairing, remove from this cell
                    remove_result rr;
                          
                    rr = ccell->remove_possibility(pos, logmsg);
                    if (rr != RR_KNOWN) {
                      advanced = true;
                      if(rr == RR_SOLVED) {
                        unsolved--;
                      }
                    }
                  }
                }
              }
            }
          }

          // Get next set of possibilities from start cell
          bool first_round = true;

          while(first_round || (num < 2 && new_set_found)) {
            // Require at least 2 possibilities marked. Also exit if
            // cannot find any more possibility sets
            first_round = false;
            new_set_found = false;
            for (i = 0; i < posarraysize && !new_set_found; i++) {
              if (!posarrayinuse[i]) {
                possibilities[posarray[i]] = true;
                posarrayinuse[i] = true;
                num++;
                new_set_found = true;
              } else {
                possibilities[posarray[i]] = false;
                posarrayinuse[i] = false;
                num--;
              }
            }
          }
        }
      }
    }
  }

  return advanced;
}


cell *sudoku::get_cell(construct_type construct, int cnum, int index)
{
  switch(construct) {
   case CT_ROW:
     return get_row_cell(cnum, index);
   case CT_COL:
     return get_col_cell(cnum, index);
   case CT_REG:
     return get_reg_cell(cnum, index);
  }

  return 0;
}


int sudoku::get_reg(construct_type construct, int cnum, int index)
{
  switch(construct) {
   case CT_ROW:
     return (cnum / base_size) * base_size + index;
   case CT_COL:
     return index * base_size + (cnum / base_size);
   case CT_REG:
     std::cerr << "Internal error! get_reg() for CT_REG!" << std::endl;
     break;
  }

  return -1;
}


cell *sudoku::get_row_cell(int row, int index)
{
  return cells[row][index];
}


cell *sudoku::get_col_cell(int col, int index)
{
  return cells[index][col];
}


cell *sudoku::get_reg_cell(int reg, int index)
{
  return cells[linear_to_y(reg, index)][linear_to_x(reg, index)];
}


cell *sudoku::get_regxy_cell(int reg, int x, int y)
{
  return cells[region_origo_y(reg) + y][region_origo_x(reg) + x];
}


int sudoku::region_origo_x(int region)
{
  return (region % base_size) * base_size;
}


int sudoku::region_origo_y(int region)
{
  return (region / base_size) * base_size;
}


int sudoku::linear_to_x(int region, int linear)
{
  return (region % base_size) * base_size + linear % base_size;
}


int sudoku::linear_to_y(int region, int linear)
{
  return (region / base_size) * base_size + linear / base_size;
}


bool sudoku::region_x(int region, int x)
{
  return x >= region_origo_x(region) &&
    x < (region_origo_x(region) + base_size);
}


bool sudoku::region_y(int region, int y)
{
  return y >= region_origo_y(region) &&
    y < (region_origo_y(region) + base_size);
}


const char *sudoku::construct_name(construct_type construct)
{
  switch(construct) {
   case CT_ROW:
     return "row";
   case CT_COL:
     return "col";
   case CT_REG:
     return "reg";
  }

  return 0;
}
